/*
  This file is a part of GDC software distributed under GNU GPL 2 licence.
  The homepage of the GDC project is http://sun.aei.polsl.pl/gdc
  
  Authors: Sebastian Deorowicz
  
  Version: 0.2
  Date   : 2011-May-16
*/

#ifndef _BIT_STREAM_H
#define _BIT_STREAM_H

#include "defs.h"
#include <iostream>
#include <vector>
#include <algorithm>
using namespace std;

typedef enum {mode_none, mode_file_read, mode_file_write, mode_file_read_ra, mode_mem_read, mode_mem_write, mode_mem_read_ra} t_mode;

// ********************************************************************************************
class CBitStream {
	FILE *file;

	unsigned char *io_buffer;
	int32 io_buffer_pos;
	int32 io_buffer_size;
	uint32 word_buffer;
	int32 word_buffer_pos;
	int32 word_buffer_size;
	t_mode mode;
	uint32 n_bit_mask[32];

public:
	inline bool FlushFullWordBuffer();
	inline bool FlushPartialWordBuffer();
	inline bool FlushInputWordBuffer();
	inline bool WriteBuffer();
	inline bool ReadBuffer();

	int32 IO_BUFFER_SIZE;
	uint64 file_pos;
	uint64 file_size;

	CBitStream();
	~CBitStream();

	// File interface
	bool Open(char *file_name);
	bool OpenRA(char *file_name);
	bool Create(char *file_name);
	bool Close();

	inline uint64 GetPos(void);
	bool SetPos(uint64 pos);

	inline bool PutBit(const uint32 word);
	inline bool Put2Bits(const uint32 word);
	inline bool PutBits(uint32 word, int32 n_bits);
	inline bool PutBytes(const unsigned char *data, int32 n_bytes);
	inline bool PutByte(const unsigned char byte);
	inline bool PutWord(const uint32 data);
	inline bool PutDWord(const uint64 data);
	inline bool Put2Bytes(const uint32 data);

	inline bool GetBit(uint32 &word);
	inline bool Get2Bits(uint32 &word);
	inline bool GetBits(uint32 &word, uint32 n_bits);
	inline bool GetBytes(unsigned char *data, uint32 n_bytes);
	inline bool GetByte(uint32 &byte);
	inline bool GetWord(uint32 &data);
	inline bool GetDWord(uint64 &data);
	inline bool Get2Bytes(uint32 &data);

	inline uint32 BitLength(const uint64 x);

	uint64 GetFilePos() {return file_pos;};
};


// ********************************************************************************************
uint64 CBitStream::GetPos(void)
{
	return file_pos;
}

// ********************************************************************************************
bool CBitStream::PutBit(const uint32 word)
{
	if(word_buffer_pos < word_buffer_size)
	{
		word_buffer <<= 1;
		word_buffer += word;
		++word_buffer_pos;
	}
	else
	{
		PutWord(word_buffer);
		word_buffer_pos = 1;
		word_buffer = word;
	}

	return true;			
};

// ********************************************************************************************
bool CBitStream::Put2Bits(const uint32 word)
{
	if(word_buffer_pos + 2 <= word_buffer_size)
	{
		word_buffer <<= 2;
		word_buffer += word;
		word_buffer_pos += 2;
	}
	else if(word_buffer_pos == word_buffer_size)
	{
		PutWord(word_buffer);
		word_buffer_pos = 2;
		word_buffer = word;
	}
	else
	{
		word_buffer <<= 1;
		word_buffer += word >> 1;
		PutWord(word_buffer);
		word_buffer = word & 1;
		word_buffer_pos = 1;
	}

	return true;			
};

// ********************************************************************************************
bool CBitStream::PutBits(uint32 word, int32 n_bits)
{
	int32 rest_bits = word_buffer_size - word_buffer_pos;
	if(n_bits >= rest_bits)
	{
		n_bits -= rest_bits;
		word_buffer <<= rest_bits;
		word_buffer += word >> n_bits;
		word &= n_bit_mask[n_bits];
		word_buffer_pos = 0;
		PutWord(word_buffer);
		word_buffer = 0;
	}
	
	word_buffer     <<= n_bits;
	word_buffer     += word;
	word_buffer_pos += n_bits;

	return true;
}

// ********************************************************************************************
bool CBitStream::FlushFullWordBuffer()
{
	PutWord(word_buffer);
	
	word_buffer     = 0;
	word_buffer_pos = 0;
	
	return true;
}

// ********************************************************************************************
bool CBitStream::FlushPartialWordBuffer()
{
	word_buffer <<= (32 - word_buffer_pos) & 7;

	if(word_buffer_pos > 24)
		PutByte(word_buffer >> 24);
	if(word_buffer_pos > 16)
		PutByte((word_buffer >> 16) & 0xFF);
	if(word_buffer_pos > 8)
		PutByte((word_buffer >> 8) & 0xFF);
	if(word_buffer_pos > 0)
		PutByte(word_buffer & 0xFF);
	
	word_buffer     = 0;
	word_buffer_pos = 0;

	return true;
}

// ********************************************************************************************
bool CBitStream::FlushInputWordBuffer()
{
	word_buffer_pos = 0;

	return true;
}

// ********************************************************************************************
bool CBitStream::WriteBuffer()
{
	fwrite(io_buffer, 1, (size_t) io_buffer_pos, file);
	io_buffer_pos = 0;

	return true;
}

// ********************************************************************************************
bool CBitStream::ReadBuffer()
{
	size_t to_read = IO_BUFFER_SIZE;

	if(to_read == 0)
		return false;

	io_buffer_size = (int32) fread(io_buffer, 1, to_read, file);
	io_buffer_pos = 0;

	return io_buffer_size > 0;
}

// ********************************************************************************************
bool CBitStream::PutByte(const unsigned char byte)
{
	if(io_buffer_pos >= IO_BUFFER_SIZE)
		WriteBuffer();
	io_buffer[io_buffer_pos++] = byte;
	file_pos++;

	return true;
}

// ********************************************************************************************
bool CBitStream::Put2Bytes(const uint32 data)
{
	PutByte((unsigned char) (data >> 8));
	PutByte((unsigned char) (data & 0xFF));

	return true;
}

// ********************************************************************************************
bool CBitStream::PutBytes(const unsigned char *data, int32 n_bytes)
{
	uint32 to_store = MIN(n_bytes, IO_BUFFER_SIZE-io_buffer_pos);
	do
	{
		copy(data, data+to_store, io_buffer+io_buffer_pos);
		io_buffer_pos += to_store;
		file_pos += to_store;
		if(io_buffer_pos >= IO_BUFFER_SIZE)
			WriteBuffer();
		if(n_bytes -= to_store)
		{
			data += to_store;
			to_store = MIN(n_bytes, IO_BUFFER_SIZE-io_buffer_pos); 
		}
		else
			to_store = 0;
	} while(to_store);

	return true;
}

// ********************************************************************************************
bool CBitStream::PutDWord(const uint64 data)
{
	PutByte(data >> 56);
	PutByte((data >> 48) & 0xFF);
	PutByte((data >> 40) & 0xFF);
	PutByte((data >> 32) & 0xFF);
	PutByte((data >> 24) & 0xFF);
	PutByte((data >> 16) & 0xFF);
	PutByte((data >> 8) & 0xFF);
	PutByte(data & 0xFF);

	return true;
}

// ********************************************************************************************
bool CBitStream::PutWord(const uint32 data)
{
	PutByte(data >> 24);
	PutByte((data >> 16) & 0xFF);
	PutByte((data >> 8) & 0xFF);
	PutByte(data & 0xFF);

	return true;
}
// ********************************************************************************************
uint32 CBitStream::BitLength(const uint64 x)
{
	for(uint32 i = 0; i < 32; ++i)
		if(x < (1ull << i))
			return i;

	return 64;
}

// ********************************************************************************************
bool CBitStream::GetBit(uint32 &word)
{
	if(word_buffer_pos == 0)
	{
		if(!GetByte(word_buffer))
			return false;
		word_buffer_pos = 7;
		word = word_buffer >> 7;
	}
	else
		word = (word_buffer >> (--word_buffer_pos)) & 1;

	return true;
};

// ********************************************************************************************
bool CBitStream::Get2Bits(uint32 &word)
{
	if(word_buffer_pos >= 2)
	{
		word = (word_buffer >> (word_buffer_pos-2)) & 3;
		word_buffer_pos -= 2;
	}
	else if(word_buffer_pos == 0)
	{
		if(!GetByte(word_buffer))
			return false;
		word = word_buffer >> 6;
		word_buffer_pos = 6;
	}
	else
	{
		word = (word_buffer & 1) << 1;
		if(!GetByte(word_buffer))
			return false;
		word += word_buffer >> 7;
		word_buffer_pos = 7;
	}

	return true;
};

// ********************************************************************************************
bool CBitStream::GetBits(uint32 &word, uint32 n_bits)
{
	word = 0;
	while(n_bits)
	{
		if(word_buffer_pos == 0)
		{
			if(!GetByte(word_buffer))
				return false;
			word_buffer_pos = 8;
		}

		if((int32) n_bits > word_buffer_pos)
		{
			word <<= word_buffer_pos;
			word += word_buffer & n_bit_mask[word_buffer_pos];
			n_bits -= word_buffer_pos;
			word_buffer_pos = 0;
		}
		else
		{
			word <<= n_bits;
			word_buffer_pos -= n_bits;
			word += (word_buffer >> word_buffer_pos) & n_bit_mask[n_bits];
			return true;
		}
	}

	return true;
}

// ********************************************************************************************
bool CBitStream::GetByte(uint32 &byte)
{
	if(io_buffer_pos >= IO_BUFFER_SIZE)
		if(!ReadBuffer())
			return false;
	byte = io_buffer[io_buffer_pos++];
	file_pos++;

	return true;
};

// ********************************************************************************************
bool CBitStream::GetBytes(unsigned char *data, uint32 n_bytes)
{
	uint32 to_read = MIN(n_bytes, (uint32) (IO_BUFFER_SIZE-io_buffer_pos));
	do
	{
		copy(io_buffer+io_buffer_pos, io_buffer+io_buffer_pos+to_read, data);
		io_buffer_pos += to_read;
		file_pos      += to_read;
		if(io_buffer_pos >= IO_BUFFER_SIZE)
			ReadBuffer();
		if(n_bytes -= to_read)
		{
			data += to_read;
			to_read = MIN(n_bytes, (uint32) (IO_BUFFER_SIZE-io_buffer_pos)); 
		}
		else
			to_read = 0;
	} while(to_read);
	
	return true;
}

// ********************************************************************************************
bool CBitStream::GetWord(uint32 &data)
{
	uint32 c;
	bool r;

	r = GetByte(c);
	data = c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;

	return r;
}

// ********************************************************************************************
bool CBitStream::GetDWord(uint64 &data)
{
	uint32 c;
	bool r;

	r = GetByte(c);
	data = c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;
	r &= GetByte(c);
	data = (data << 8) + c;

	return r;
}

// ********************************************************************************************
bool CBitStream::Get2Bytes(uint32 &data)
{
	uint32 c;
	bool r;

	r = GetByte(c);
	data = c;
	r &= GetByte(c);
	data = (data << 8) + c;

	return r;
}

#endif
